// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

// This file gathers together all of the pieces of the 2-3 Tree implementation

#include "dictionary.h"

// Empty key value
int EMPTYKEY = -1;

#include "ttnode.h"

#include "tttree.h"

// Print out a 2-3 Tree
template <typename Key, typename Elem>
void TTTree<Key, Elem>::
printhelp(TTNode<Key,Elem>* root, int level) const {
  if (root == NULL) return;
  for (int i=0; i<level; i++) cout << "   "; // indent
  cout << root->lval;
  if (root->rkey != EMPTYKEY)
    cout << "  " << root->rval << "\n";
  else cout << "\n";
  printhelp(root->left, level+1);
  printhelp(root->center, level+1);
  if (root->rkey != EMPTYKEY)
    printhelp(root->right, level+1);
}

// Insert a record into the 2-3 tree. "Root" is the current
// node, "e" is the Elem to be inserted, "retval" and
// "retptr" are return values for a promoted Elem and child
// node when the current node is split. 
template <typename Key, typename Elem>
void TTTree<Key, Elem>::
inserthelp(TTNode<Key,Elem>*& root, Key K, Elem it,
           Key& retkey, Elem& retval,
           TTNode<Key,Elem>*& retptr) {
  Elem myretv;       // Local copies: return record
  Key myretk;        //               return key
  TTNode<Key,Elem>* myretp = NULL; // return subkey
  if (root == NULL) {        // Empty tree: make new node
    root = new TTNode<Key,Elem>();
    root->lval = it; root->lkey = K;
  }
  else if (root->isLeaf())   // At leaf node: insert here
    if (root->rkey == EMPTYKEY) { // Only one node stored
      if (K > root->lkey) {       // Put it to right
        root->rval = it;
        root->rkey = K;
      }
      else {                      // Put it to left
        root->rval = root->lval;
	root->rkey = root->lkey;
        root->lval = it;
	root->lkey = K;
      }
    }
    else splitnode(root, K, it, NULL, retkey,
                   retval, retptr);
  else if (K < root->lkey)  // Process internal node
    inserthelp(root->left, K, it, myretk, myretv, myretp);
  else if ((root->rkey == EMPTYKEY) || (K < root->rkey))
    inserthelp(root->center, K, it, myretk, myretv, myretp);
  else
    inserthelp(root->right, K, it, myretk, myretv, myretp);
  if (myretp != NULL) // Child split: receive promoted value
  {
    if (root->rkey != EMPTYKEY) // Full: split
      splitnode(root, myretk, myretv, myretp,
                retkey, retval, retptr);
    else { // Not full: add to this node
      if (myretk < root->lkey) { // Put on left side
        root->rval = root->lval;
        root->rkey = root->lkey;
        root->lval = myretv;
	root->lkey = myretk;
        root->right = root->center;
        root->center = myretp;
      }
      else { // Put on right side
        root->rval = myretv;
        root->rkey = myretk;
        root->right = myretp;
      }
    }
  }
}

// Split a node in 2-3 tree.  "Root" is node being split,
// "K" is the key being inserted, "it" is the Elem being
// inserted, "inptr" is the child node being added (if root
// is an internal node), "retkey", "retval" and "retptr"
// are the promoted key, value and child, respectively.
template <typename Key, typename Elem>
void TTTree<Key,Elem>::splitnode(TTNode<Key,Elem>* root,
	    Key K, Elem it, TTNode<Key,Elem>*inptr,
            Key& retkey, Elem& retval,
            TTNode<Key,Elem>*& retptr) {
  retptr = new TTNode<Key,Elem>();  // Node created by split
  if (K < root->lkey) {      // Add at left
    retval = root->lval;
    retkey = root->lkey;
    root->lval = it;
    retptr->lval = root->rval;
    retptr->lkey = root->rkey;
    retptr->left = root->center;
    retptr->center = root->right;
    root->center = inptr;
  }
  else if (K < root->rkey) { // Add at center
    retval = it;
    retkey = K;
    retptr->lval = root->rval;
    retptr->lkey = root->rkey;
    retptr->left = inptr;
    retptr->center = root->right;
  }
  else {                     // Add at right
    retval = root->rval;
    retkey = root->rkey;
    retptr->lval = it;
    retptr->lkey = K;
    retptr->left = root->right;
    retptr->center = inptr;
  }
  root->rkey = EMPTYKEY;
}

// Find the record that matches a given key value
template <typename Key, typename Elem>
bool TTTree<Key, Elem>::
findhelp(TTNode<Key,Elem>* root, Key K, Elem& it) const {
  if (root == NULL) return false;    // value not found
  if (K == root->lkey)
    { it = root->lval; return true; }       // Found it
  if (K == root->rkey)
    { it = root->rval; return true; }       // Found it
  if (K < root->lkey)                       // Go left
    return findhelp(root->left, K, it);
  else if (root->rkey == EMPTYKEY)          // 2 child node
    return findhelp(root->center, K, it);   // Go center
  else if (K < root->rkey)
    return findhelp(root->center, K, it);   // Go center
  else return findhelp(root->right, K, it); // Go right
}
